#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "zwordcount.h"
#include "zsearcher.h"

/* ****************************************************************************
 * zsearcher
 *
 * Manages word_count, running_word_count, and hittimes.
 *
 * Calls searchAndAdd to put words in their place in a binary sorted tree,
 * creating new nodes as needed. 
 *
 * Creates two identical doubly linked lists *in the order received*, one from
 * dummyhead-> next/prev -> next/prev -> ..., and one from
 * dummyhead-> fwrd/back -> fwrd/back -> .... 
 * The next/prev is desired, and kept, because one use of the data requires it,
 * and the best place to make that list is here, as the words arrive.
 * The fwrd/back list is there to be sorted in order of word frequency. I just
 * didn't want to destroy the other list in doing so.
 *
 * So the struct becomes is a binary sorted tree, with two independently 
 * organized linked lists running through as well.
 * ************************************************************************** */

node* ptr2root ;
node* lastNew ;   // free-floating pointer to build a linked list.
node* dummyhead ; // dummy head for the head of the linked list.
node* dummytail ; // dummy tail for the tail of the linked list.

node* searchAndAdd (node* anode, node* last, char* incoming_word) ;

void putwordhere(char* incoming_word, node* ptr2root) ;

/* ************************************************************************** */
node* zsearcherInit ()
{
  dummyhead  = (node*) malloc(sizeof(node)) ;
  if (dummyhead  == NULL) { printf("\nCan't malloc a new word_struct. Bailing\n"); exit (1) ; }
  dummyhead->valid = 0 ; 

  dummytail  = (node*) malloc(sizeof(node)) ;
  if (dummytail  == NULL) { printf("\nCan't malloc a new word_struct. Bailing\n"); exit (1) ; }
  dummytail->valid = 0 ; 

  ptr2root  = (node*) malloc(sizeof(node)) ;
  if (ptr2root  == NULL) { printf("\nCan't malloc a new word_struct. Bailing\n"); exit (1) ; }
  putwordhere("notAword", ptr2root) ;
  ptr2root->valid = 0 ; 

  /* There are only two non-valid nodes, not valid for words that is.
   * ptr2root does not have a valid word, because all words are converted to lower case. 
   * But it starts with an 'n', so it's here to get the tree off to a balanced
   * start. When it comes time to sort the data it will be necessary to test for this, 
   * so I've also added a valid bit, quicker to test than a string compare. */

  ptr2root->next  = dummyhead ;
  ptr2root->fwrd  = dummyhead ;
  dummyhead->prev = ptr2root ; 
  dummyhead->back = ptr2root ; 
  dummytail->next = dummytail ; 
  dummytail->fwrd = dummytail ; 

  lastNew = dummyhead ; 

  return (ptr2root) ;
}



/* ****************************************************************************
 * shuttles the incoming word from the caller to the searchAndAdd(node)
 * function then fills out the node.
 * ************************************************************************** */
void gotWord (char *incoming_word) 
{
  static int wordcount = 0 ; // Running word count, not individual word count
  node *current_ptr ; 
  int   i ;
  int*  ip_tmp ;

  current_ptr = searchAndAdd(ptr2root, ptr2root, incoming_word) ; 
  wordcount++ ;

  if (current_ptr->count == 0)  // a new node
  {
    putwordhere(incoming_word, current_ptr) ;
    current_ptr->count = 1 ;

    current_ptr->hittimes = (int*) malloc((1)*sizeof(int)); 
    if (current_ptr->hittimes == NULL) { printf("\nCan't malloc for 1 bloody int! Bailing 16.\n") ; exit (1) ; }
    current_ptr->hittimes[0] = wordcount ;

    pushWSLNW(wordcount, 0, 0) ;
  }
  else // an existing node
  {
    current_ptr->count++ ;

    ip_tmp = NULL ;
    ip_tmp = (int*) malloc((current_ptr->count)*sizeof(int)); 
    if (ip_tmp == NULL) { printf("\nCan't malloc for %d ints!. Bailing 20.\n", current_ptr->count) ; exit (1) ; }
    for (i=0 ; i < (current_ptr->count - 1) ; i++) 
    {
       ip_tmp[i] = current_ptr->hittimes[i] ;
    }
    free(current_ptr->hittimes) ; 

    ip_tmp[current_ptr->count-1] = wordcount ;
    current_ptr->hittimes = ip_tmp ;
  }
}



/* ****************************************************************************
 * Traverse the tree 
 * If it finds a match on the incoming_word, it returns a pointer to that node.
 * If it does not find a match on the incoming_word, it 
 * -- adds a node to the tree
 * -- makes count = 0,
 * -- initializes other pointers, and
 * -- returns a pointer to that node.
 * ************************************************************************** */
node* searchAndAdd (node* anode, node* last, char * incoming_word)
{
  node* newnode ;

  if( anode == NULL) 
  {
    newnode = (node*) malloc(sizeof(node)) ;
    if (newnode  == NULL) { printf("\nCan't malloc a new word_struct for one more node! Bailing.\n"); exit (1) ; }

    newnode->left  = 0 ;
    newnode->right = 0 ;

    if (strcmp(incoming_word, last->word)<0) { last->left  = newnode ; }
    else                                     { last->right = newnode ; }

    newnode->prev = lastNew ;
    newnode->back = lastNew ;
    lastNew->next = newnode ;
    lastNew->fwrd = newnode ;
    newnode->next = dummytail ;
    newnode->fwrd = dummytail ;

    lastNew = newnode ; 

    newnode->count = 0 ;

    return newnode; 
  }
  else // this node already exists, with content
  {
    if (strcmp(incoming_word,anode->word)==0)
    {
      return anode ; // That is, do nothing. The caller will update the counts
    }
    else if (strcmp(incoming_word,anode->word)<0)  
    {
      return searchAndAdd(anode->left,  anode, incoming_word);
    }
    else
    {
      return searchAndAdd(anode->right, anode, incoming_word);
    }
  }
  printf ("\nBailing. This should not be happening.") ; exit (0) ;
  return anode ; // should never happen. FIXME and verify that
}



/* ************************************************************************** */
void putwordhere(char* word, node* here) 
{
  here->word = (char *) malloc(strlen(word)+1) ;
  if (here->word == NULL) { printf("\nCan't malloc for %s @ %p. putwordhere bailing.\n", word, here) ; exit (1) ; }
  strcpy(here->word, word) ;
}


