/*****************************************************************--701234567890
 *                   Dan Downs  OstensiblyMyID@gmail.com
 * Zipf's Law program.
 * Version 1.0 will simply create a word frequency list, sorted most to least.
 * Version 2.0 adding ...
 *     a) option to include digits in/as words
 * Version 3.0
 *     command line options for making more sensible csv files for gnuplot
 * Version 4.0
 *     Split into ztokenizer.c, zsearcher.c, and zwordcount.c
 *     ztokenizer:
 *        Identifies words in the file passed to it, calls zsearcher.
 *     zsearcher:
 *        Puts words in their place, does the accounting.
 *     zwordcount:
 *        Responsible for all opening and closing of files.
 *        Owns the top level management of the search tree.
 *        Prints results.

 * Below is a long word. It is 5 alphabets long. It is here to make certain the 
 * BIG word detector and truncator works buy running this program on its own
 * source code. abcedfghijklmnopqrstuvwxyzabcedfghijklmnopqrstuvwxyzabcedfghijklmnopqrstuvwxyzabcedfghijklmnopqrstuvwxyzabcedfghijklmnopqrstuvwxyz
 * I run this program on itself because with all the special characters, it's a
 * rather abrassive test for the character-by-character processing.
 *
 * Expecting that the volume of data this program will accumulate is far beyond
 * anything I want to manage, this program spills everything to a single output
 * stream, each line potentially prefixed with a KEY. Any script will be able to
 * extract the data according to the needs at hand. Here are the line prefixes
 * and their meaning:
 *  WCvsRANK <word> <word's rank> <word's count> sorted by rank, highest first.
 *  WSLNW <wordcount> <words since last new>
 *    where <wordcount> is the current wordcount of all words read in
 *    where <words since last new> number of words since a new word was found and added
 *    (this is independent of the actual word)
 *  WRPROF <word> <list of numbers>
 *    where <word-in-question> is the word in question, and
 *    where <list of numbers> are the wordcounts where the word-in-question occurred
 *    (this is about a particular word)
 *  # QFACT: <stuff>, where <stuff> is any single statistic I want to call out
 * ************************************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "zwordcount.h"
#include "ztokenizer.h"
#include "zsearcher.h"

// The only input file
FILE *fp_R ;

// Output files for gnuplot.
// WCvsRANK is word count versus word rank.
// WSLNW    means words gone by since the last time a new word arrived
// WRPROF   stands for Word Rank Profile
FILE *fp_WCvsRANKdat ;
FILE *fp_WCvsRANKcsv ;
FILE *fp_WSLNWdat ;
FILE *fp_WSLNWcsv ;
FILE *fp_WRPROFdat ;
FILE *fp_WRPROFcsv ;

void Usage  (void) ;
void Info   (void) ;

char* createfilename ( char* infilename, char* targetdir, const char* therest) ;

void pushWRPROF                 (node* current, char* infilename, int finish) ;
void PrintWordFrequencyList     (node* head,    char* infilename, int opt_cols, int opt_all) ;
void PrintWordFrequencyTree     (node* root) ;
node* BuildSortedListFromTree   (node* root) ;



/* *****************************************************************************
 *                                  MAIN
 * ************************************************************************** */
int main (int argc, char *argv[])
{

  int i, j, l ;

  int opt_all    = 0 ; /* For creating and writing the output files, else STDOUT. */
  int opt_cols   = 0 ; /* print just the 2 columns, rank and count. */
  int opt_odir   = 0 ; /* print to an output directory, else print to this dir. */
  int opt_split  = 0 ; /* For a single hyphenated word, do not preserve the hyphen if "split" is set*/
  int opt_digits = 0 ; /* include as characters @, [, `, {, /, : and 0-9. */

  int getWords_retval = 1 ;

  char   inchar ;

  const char* WCvsRANKdat = ".WCvsRANK.dat" ;
  const char* WCvsRANKcsv = ".WCvsRANK.csv" ;
  const char* WSLNWdat    = ".WSLNW.dat" ;
  const char* WSLNWcsv    = ".WSLNW.csv" ;
  const char* WRPROFdat   = ".WRPROF.dat" ;
  const char* WRPROFcsv   = ".WRPROF.csv" ;

  char* infile ;
  char* infilename ;
  char* nameWCvsRANKdat ;
  char* nameWCvsRANKcsv ;
  char* nameWSLNWdat ;
  char* nameWSLNWcsv ;
  char* nameWRPROFdat ;
  char* nameWRPROFcsv ;
  char* outdir ;

  node* root ;    /* pointer to the binary tree root */
  node* head ;    /* pointer to a first element for both linked lists */

  root = zsearcherInit() ; 

  /* 
   * Parse command line
   */
  for (i=1;i<argc;i++)      {
    if (argv[i][0] == '-')  {
      switch (argv[i][1])   {
        case 'i': Info() ;
        case 'h': Usage() ;
        case 'c': opt_cols   = 1 ; break ;
        case 's': opt_split  = 1 ; break ;
        case 'd': opt_digits = 1 ; break ;
        case 'a': opt_all    = 1 ; break ;
  
        case 'o': opt_odir   = 1 ; opt_all = 1 ;
                  i++ ;
                  outdir = (char*) malloc((strlen(argv[i])+1)*sizeof(char)) ;
                  if (outdir == NULL) { printf("\nCan't malloc for outdir name. Bailing.\n") ; exit (1);}
                  strcpy(outdir, argv[i]) ;
                  break ;
  
        case 'f': i++ ;
                  infile = (char*) malloc((strlen(argv[i])+1)*sizeof(char)) ;
                  if (infile == NULL) { printf("\nCan't malloc for filename string. Bailing.\n") ; exit (1);}
                  strcpy(infile,argv[i] ) ;
                  fp_R = fopen(infile,"r") ;
                  if (fp_R == NULL) { printf("\nCan't open the file \"%s\".\nBailing.\n", infile) ; exit(1) ; }
                  break ;
  } } }

  /* ********************************************************************
   *  Housekeeping ... check command line options
   */
  if (fp_R == NULL)    { fprintf(stderr, "\nNo -f <filename> specified. Bailing 8.\n") ;    exit (1) ; }
  if (opt_digits == 1) { printf("\nIncluding digits in words.\n") ; }
  if (opt_odir == 0)   { outdir = NULL ; }
  
  /* -------------------------------------------------------------
   * This option implies creating and writing the output files.
   */
  if (opt_all == 1)
  {
    /* ---------- extract "infilename" from infile ---------- */
    l = strlen(infile)  ;
    i = l  ;
    inchar = infile[i] ; // the final character.
  
    while ((i>0) && (inchar != '/')) { inchar = infile[--i] ; } // work backwards to the '/'

    if (inchar == '/') { i++ ; }

    infilename = (char*) malloc (( l - i +1 ) * sizeof(char)) ;
    j = 0 ;
    while (i < (l +1)) { infilename[j++] = infile[i++] ; }

    infilename[--j] = '\0' ;
    /* ------------------------------------------------------- */

    /* Create filenames for the output files using filename */
    nameWCvsRANKdat = createfilename (infilename, outdir, WCvsRANKdat) ; fp_WCvsRANKdat = fopen(nameWCvsRANKdat, "w") ;
    if (fp_WCvsRANKdat == NULL) {printf("\nCan't create file %s%s Bailing 9.\n", infilename, WCvsRANKdat) ; exit(1) ; }

    nameWCvsRANKcsv = createfilename (infilename, outdir, WCvsRANKcsv) ; fp_WCvsRANKcsv = fopen(nameWCvsRANKcsv, "w") ;
    if (fp_WCvsRANKdat == NULL) {printf("\nCan't create file %s%s Bailing 10.\n", infilename, WCvsRANKcsv) ; exit(1) ; }

    nameWSLNWdat    = createfilename (infilename, outdir, WSLNWdat) ; fp_WSLNWdat    = fopen(nameWSLNWdat, "w") ;
    if (fp_WSLNWdat == NULL) {printf("\nCan't create file %s%s Bailing 11.\n", infilename, WSLNWdat) ;    exit(1) ; }

    nameWSLNWcsv    = createfilename (infilename, outdir, WSLNWcsv) ; fp_WSLNWcsv    = fopen(nameWSLNWcsv, "w") ;
    if (fp_WSLNWcsv == NULL) {printf("\nCan't create file %s%s Bailing 12.\n", infilename, WSLNWcsv) ;    exit(1) ; }

    nameWRPROFdat   = createfilename (infilename, outdir, WRPROFdat) ; fp_WRPROFdat   = fopen(nameWRPROFdat, "w") ;
    if (fp_WRPROFdat == NULL) {printf("\nCan't create file %s%s Bailing 13.\n", infilename, WRPROFdat) ;   exit(1) ; }

    nameWRPROFcsv   = createfilename (infilename, outdir, WRPROFcsv) ; fp_WRPROFcsv   = fopen(nameWRPROFcsv, "w") ;
    if (fp_WRPROFcsv == NULL) {printf("\nCan't create file %s%s Bailing 14.\n", infilename, WRPROFcsv) ;   exit(1) ; }
  
    pushWSLNW  (0, 0, infilename) ;
  }
  /* end of if (opt_all == 1) and end of housekeeping
   * ******************************************************************** */




  /* **************************************************************************
   * This next code is a manager and knows nothing about what is going on.
   * Everything is set and ready to go, so now it: 
   * 1) calls the tokenizer getWords,
   *    getWords passes that via gotWord where it is inserted into the tree,
   * 2) then when getWords returns, takes all the credit by printing what's been 
   *    created. 
   * ************************************************************************ */
  getWords_retval = getWords(fp_R, opt_split, opt_digits, opt_all);
				
  if (getWords_retval == 1) { printf("\ngetWords_retval == 1 ... Tragic mishap!\n") ; }
  fclose (fp_R) ;

  // For testing only
  //PrintWordFrequencyTree ( root ) ;

  head = BuildSortedListFromTree ( root ) ;

  PrintWordFrequencyList (
    head, 
    infilename,
    opt_cols,
    opt_all
  ) ;

  if (opt_all == 1)
  {
    fclose (fp_WCvsRANKdat) ;
    fclose (fp_WCvsRANKcsv) ;
    fclose (fp_WSLNWdat) ;
    fclose (fp_WSLNWcsv) ;
    fclose (fp_WRPROFdat) ;
    fclose (fp_WRPROFcsv) ;

    free (nameWCvsRANKdat) ;
    free (nameWCvsRANKcsv) ;
    free (nameWSLNWdat) ;
    free (nameWSLNWcsv) ;
    free (nameWRPROFdat) ;
    free (nameWRPROFcsv) ;

    free (infilename) ;
  }

  free (infile) ;
  free (outdir) ;
  leave(root) ;
  if (head != NULL) { free (head) ;}
  return (0) ; 
}




/* ****************************************************************************
 * Up until now, the back and fwrd pointers in nodes have been untouched, save
 * for being a copy of the next and prev pointers.
 * This routine walks the tree to build a doubly linked list within the
 * tree sorted from most frequent word to least. For entries with matching
 * counts, the word with the first occurence will be put closer to the
 * head of the list. 
 * ************************************************************************** */

node* BuildSortedListFromTree (node * anode ) 
{
  node* iptr ;
  node* jptr ;
  node* t ;

  // The root node->fwrd points to the dummy node. dummy->fwrd points to the first real node
  iptr = anode->fwrd->fwrd ;
  jptr = iptr->fwrd ;

  while (iptr->fwrd != iptr)
  {
    while (jptr->fwrd != jptr)
    {
      if (jptr->count > iptr->count)
      {
        t = jptr ;

        // splice out t
        t->back->fwrd = t->fwrd ;
        t->fwrd->back = t->back ;

        // splice in detached t node ahead of iptr node 
        t->fwrd = iptr ;
        t->back = iptr->back ;
        iptr->back->fwrd = t ;
        iptr->back       = t ;

        iptr = t ;
      }
      jptr = jptr->fwrd ;
    }
    iptr = iptr->fwrd ;
    jptr = iptr->fwrd ;
  }
 
  return anode->fwrd->fwrd  ;
}

/* --------------------------------------------------------------------
 * Just for testing ... walk the tree inorder and print the results
 */
void PrintWordFrequencyTree ( node* anode )
{
  if (anode->left)  PrintWordFrequencyTree(anode->left);
  printf("%d : %s\n",anode->count, anode->word);
  if (anode->right) PrintWordFrequencyTree(anode->right);
}





/* --------------------------------------------------------------------
 * mallocs, copys, and cats to splice all the args into a filename
 */
char* createfilename (
   char* infilename,
   char* targetdir, 
   const char* therest
   )
{
  char* thefilename ;

  if (targetdir == NULL) {
     thefilename = (char*) malloc ((                    strlen(infilename) + strlen(therest) + 1)*sizeof(char)) ;
  } 
  else {
     thefilename = (char*) malloc ((strlen(targetdir) + strlen(infilename) + strlen(therest) + 2)*sizeof(char)) ;
  }

  if (thefilename == NULL) { printf("\nCan't malloc for %s. Bailing 15.\n", thefilename) ; exit (1);}

  if (targetdir != NULL) {
    strcpy(thefilename, targetdir) ;
    strcat(thefilename, "/") ;
    strcat(thefilename, infilename) ;
  }
  else {
    strcpy(thefilename, infilename) ;
  }
  strcat(thefilename, therest) ;

  return thefilename ;
}




/* -----------------------------------------------
 * Free all and exit. 
 * Postorder traversal, but free the word and free the hittimes 
 * before freeing the node. 
 */
void leave ( node* anode )
{
  if ( anode == NULL ) return ;
  leave(anode->left) ;
  leave(anode->right) ;
  free (anode->word) ;
  free (anode->hittimes) ;
  free (anode) ;
}

/* -----------------------------------------------
 * print out the particular word's repition profile 
 */
void pushWRPROF (node* current, char* infilename, int finish)
{
         int i ;
         int xmin_WRPROF ;
  static int xmax_WRPROF ;
         int ymin_WRPROF ;
  static int ymax_WRPROF ;

  if (finish == 0) 
  {
    fprintf(fp_WRPROFcsv, "%d %d %s\n", current->hittimes[0], current->hittimes[0], current->word) ;
    for (i=1;i < current->count; i++)
    {
      fprintf(fp_WRPROFcsv, "%d %d\n", current->hittimes[0], current->hittimes[i]) ;
    }
  }
  else 
  {
    xmin_WRPROF = 0 ;
    ymin_WRPROF = 0 ;
    xmax_WRPROF = finish ;
    ymax_WRPROF = finish ;
    fprintf (fp_WRPROFdat, "set title \"File: %s : Word Repetition Profile\" ; set xrange [%d : %d] ; set yrange [%d : %d] ; set terminal png ; set output \"%s.WRPROF.png\" ; plot \"%s.WRPROF.csv\" w points lc 0 ; set terminal x11 ; replot", infilename, xmin_WRPROF, xmax_WRPROF, ymin_WRPROF, ymax_WRPROF, infilename, infilename) ;
  }
}




/* ----------------------------------------------------------------------------
 *      Print the list pointed to, with stats.
 *  This routine wants to walk word_struct as a singly linked list from head
 *  to tail, expecting data to be sorted by frequency, highest at the head to 
 *  lowest at the tail. For any entries with identical frequency, sort those by 
 *  word occurence, earliest to the head. 
 */
void PrintWordFrequencyList  (
   node* head,
   char* infilename,
   int opt_cols,
   int opt_all
   )
{
  node *current ;
  int rank ; 
  int totalcount ;
  int uniquewords ;
  int runningcount ;
  int usedonlyonce ;
  int previouswordcount ;
  int lessThanHalfVolume ;

  float uniqRATIOtotal ;
  float runningOVERtotal ;
  float averageOfRatios ;

  int xmin_WCvsRANK = 1 ;
  int xmax_WCvsRANK = 1 ;
  int ymin_WCvsRANK = 1 ;
  int ymax_WCvsRANK = 1 ;

  totalcount = 0 ;
  uniquewords = 0 ; // same as max value of rank
  previouswordcount = 0 ;
  averageOfRatios = 0.0 ;

  current = head ;

  printf("\n") ;

  // --------------------------------------------------------------------------------------
  // This section runs through the list only to collect the stats.
  // totalcount ,uniquewords, uniqRATIOtotal, and averageOfRatios
  while ( current->fwrd != current )
  {
    if (current->word != NULL) 
    {
      ++uniquewords ;
      totalcount += current->count ;
    }
    if (current == head) 
    {
      previouswordcount = current->count ;
    }
    else
    {
      averageOfRatios = averageOfRatios + (float)(previouswordcount)/(float)(current->count) ;
      previouswordcount = current->count ;
    }
    current = current->fwrd ;
  }


  if (totalcount > 0.0) 
  {
    uniqRATIOtotal = (float)uniquewords / (float)totalcount ;
  }
  averageOfRatios = averageOfRatios/((float)(uniquewords - 1)) ;

  // This just prints the header info.
  printf("\n# QFACT: The total word count is \t%d",totalcount) ;
  printf("\n# QFACT: The unique words number \t%d",uniquewords) ;
  printf("\n# QFACT: The ratio unique/total is \t%6.4f %%", 100*uniqRATIOtotal) ;
  printf("\n# -------------------------------------------------------------------------") ;
  printf("\n#     \tword \tthe                                               " ) ;
  printf("\n# rank\tcount \tword                                              " ) ;
  printf("\n# -------------------------------------------------------------------------") ;

  rank = 1 ;
  usedonlyonce = 0 ;
  runningcount = 0 ;
  lessThanHalfVolume = 0 ;
  current = head ;

  // --------------------------------------------------------------------------------------
  // Yes, this is a second pass through the list.
  while ( current->fwrd != current )
  {
      if (opt_all == 1)
      {
         pushWRPROF (current, infilename, 0) ;
      }
      runningcount += current->count ;
      runningOVERtotal = (float)runningcount / (float)totalcount ;
      if ((lessThanHalfVolume == 0) && (runningOVERtotal >= 0.5 ))
      {
         printf("\n# QFACT: The highest ranking %d words make up %6.3f %% of the total, OR", rank, 100* runningOVERtotal ) ;
         printf("\n# QFACT: %6.3f %% of the unique words make up %6.3f %% of the total volume", 100* (float)rank/(float)uniquewords, 100* runningOVERtotal ) ;
         lessThanHalfVolume = 1 ;
      }
      //previous_count = current->count ;

      if (opt_all == 1)
      {
         fprintf(fp_WCvsRANKcsv, "%d %d\n",rank, current->count) ;
      }
      
      printf("\n%d, \t%d",rank,current->count) ;
      if (opt_cols == 0) //print just the 2 column
      {
         printf(", \t%s, \t%d",current->word, current->hittimes[0]) ;
      }

      if (current->count == 1) ++usedonlyonce ;
      ++rank ;
      current = current->fwrd ;
  }
  printf("\n# QFACT: The number of words used only once \t%d", usedonlyonce) ;
  printf("\n# QFACT: The ratio (words used only once)/total is \t%6.4f %%", (float)(100*usedonlyonce)/(float)totalcount) ;
  printf("\n# QFACT: The average of the ratios between frequencies is \t%6.4f", averageOfRatios) ;
  printf("\n") ;

   xmax_WCvsRANK = uniquewords ;
   ymax_WCvsRANK = head->count ;

   if (opt_all == 1) // write out all .dat and .csv files
   {
      pushWSLNW  (totalcount, 1, 0) ;          //finish. This time pass totalcount, not the current wordcount.
      pushWRPROF (current,    infilename, totalcount) ; //finish. Pass totalcount (sb > 0) to finish up.

     fprintf (fp_WCvsRANKdat, "set title \"File: %s : log(word count) vs log10(word rank)\" ; set xrange [%d : %d] ; set yrange [%d : %d] ; set logscale y ; set logscale x ; set terminal png ; set output \"%s.WCvsRANK.png\" ; plot \"%s.WCvsRANK.csv\" w lines lc 3 ; set terminal x11 ; replot", infilename, xmin_WCvsRANK, xmax_WCvsRANK, ymin_WCvsRANK, ymax_WCvsRANK, infilename, infilename) ;
   }
}


/* ------------------------------------------------------------------------------
 * Independent of the words counted, this will give a rate of new word production.
 * Called each time a new word is found. Call with finish == 0 to make a new entry
 * in the file fp_WSLNWcsv. Call with finish == 1 to make the .dat file.
 */
void pushWSLNW (int wordcount, int finish, char* filename)
{
         int xmax_WSLNW ;
  static int ymax_WSLNW = 0 ;
  static int wslnw = 0 ;
  static char* infilename ;

  if (filename != NULL) 
  {
    infilename = filename ; // One time only
  }
  else if (infilename != NULL)
  {
    if (finish == 0) 
    {
      fprintf(fp_WSLNWcsv, "%d %d\n",wordcount, wordcount-1-wslnw) ;
      if (wordcount-1-wslnw > ymax_WSLNW)
      {
        ymax_WSLNW = wordcount-1-wslnw ;
      }
      wslnw = wordcount-1 ;
     }
     else 
     {
       xmax_WSLNW = wordcount + 1 ; //This time, "wordcount" should be from totalcount.
       ymax_WSLNW += 1 ;
       fprintf(fp_WSLNWdat, "set title \"File: %s : log10(word count) vs (WSLNW)\" ; set xrange [0.1 : %d] ; set yrange [0.1 : %d] ; set logscale x ; set terminal png ; set output \"%s.WSLNW.png\" ; set pointsize -1 ; plot \"%s.WSLNW.csv\" w impulses lc 2; set terminal x11 ; replot", infilename, xmax_WSLNW, ymax_WSLNW, infilename, infilename) ;
     }
  }
}


/* -------------------------------------------------------------
 * Just the usage
 */
void Usage (void) {
   printf("\nUsage:") ;
   printf("\n\tzipfs\t[-h|-i] [-d] [-c] [-a] [-s]") ;
   printf("\n\t \t[-o <destination_dir>] -f [path/]<inputfile>") ;
   printf("\nwhere:") ;

   printf("\n-f <inputfile> is the name of the text file to read. A path can be\n  specified with the filename <inputfile>. With some options, several\n  ouput files are created and the filename is used as their base name.\n") ;

   printf("\n-o <output_dir> for the .dat and .csv files. Selecting -o activates\n  -a as well. With some options, and in particular when used with the\n  zwc_gui.tcl front end, many ouput files are created. Therefore it\n  is highly recommended to create and use a separate output directory.\n") ;

   printf("\n-s (split) means that hyphenated words will not be considered as\n  single words, EXCEPT where the hyphen is at eol. For example, with\n  -s, the  text: \n\t \"... doorknob door- \n\t knob work-bench ...\"  \n  will be considered to contain three words, \"doorknob\", \"work\", and\n  \"bench\", but without the -s option, there will be two words,\n  \"doorknob\" and \"workbench\" and the count for \"doorknob\" will be 2.\n") ;

   printf("\n-h is this help, -i will dump out details of what this program does,") ;

   printf("\n-d to include digits in/as words. Default is not to include digits,") ;

   printf("\n-c to output 2 columns, \"rank, count\" (default is \"rank, count, word\"),") ; 

   printf("\n-a (overrides -c) will output the following seven items:") ;
   printf("\n\t output to STDOUT, plus three pairs of files for GNUplot:") ;
   printf("\n\t inputfile.WCvsRANK.dat\tand inputfile.WCvsRANK.csv") ;
   printf("\n\t inputfile.WSLNW.dat \tand inputfile.WSLNW.csv") ;
   printf("\n\t inputfile.WRPROF.dat \tand inputfile.WRPROF.csv") ;
   printf("\n\n") ;
   exit (0);
}





/* *********************************************************************************
 * Extended info 
 */
void Info (void)
{
  printf("\nThis program reads text files and outputs a word list, sorted from most to");
  printf("\nleast used words. It also prints some statistics/factoids, which are:");
  printf("\n-- words sorted by rank, where the highest ranking word is the most used,");
  printf("\n-- unique words vs total, where \"unique words\" is the number of ranks,") ;
  printf("\n-- number of unique words accounting for 50+percent of the total words,") ;
  printf("\n-- number of words used only once vs the total number of words.") ;
  printf("\nThe -a option creates a number of extra files intended to be used with") ;
  printf("\ngnuplot and the companion zwc_gui.tcl script to generate plots. Without") ;
  printf("\nthe zwc_gui.tcl front end the plots can still be generated from the ") ;
  printf("\n<output_dir> with the commands:\n") ;
  printf("\n\tgnuplot -background white -persist <inputfile>.WCvsRANK.dat\n") ;
  printf("\n\tgnuplot -background white -persist <inputfile>.WSLNW.dat\n") ;
  printf("\n\tgnuplot -background white -persist <inputfile>.WRPROF.dat\n") ;
  printf("\n\n") ;
  exit (0);
}


